package main

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"os"
	"time"

	"github.com/go-ping/ping"
	"sneak.berlin/go/mullvadclosest"
)

// Relay represents the structure of each relay in the JSON file
type Relay struct {
	Hostname         string      `json:"hostname"`
	Ipv4AddrIn       string      `json:"ipv4_addr_in"`
	Ipv6AddrIn       string      `json:"ipv6_addr_in"`
	IncludeInCountry bool        `json:"include_in_country"`
	Active           bool        `json:"active"`
	Owned            bool        `json:"owned"`
	Provider         string      `json:"provider"`
	Weight           int         `json:"weight"`
	EndpointData     interface{} `json:"endpoint_data"` // Can be a string (e.g., "openvpn") or a nested struct
	Location         struct {
		Country     string  `json:"country"`
		CountryCode string  `json:"country_code"`
		City        string  `json:"city"`
		CityCode    string  `json:"city_code"`
		Latitude    float64 `json:"latitude"`
		Longitude   float64 `json:"longitude"`
	} `json:"location"`
}

// String returns a string representation of the Relay
func (r Relay) String() string {
	return fmt.Sprintf("Relay(hostname=%s, ip=%s, country=%s)", r.Hostname, r.Ipv4AddrIn, r.Location.Country)
}

// CheckLiveness checks if the relay is live
func (r Relay) CheckLiveness() (bool, time.Duration, error) {
	pinger, err := ping.NewPinger(r.Ipv4AddrIn)
	if err != nil {
		return false, 0, err
	}
	pinger.Count = 1
	pinger.Timeout = 3 * time.Second // Increased timeout for the single ping
	pinger.SetPrivileged(true)
	pinger.Run()
	stats := pinger.Statistics()
	if stats.PacketsRecv == 0 {
		return false, 0, nil
	}
	return true, stats.AvgRtt, nil
}

// MeasureLatency measures the minimum latency of the relay over 5 pings
func (r Relay) MeasureLatency() (time.Duration, error) {
	pinger, err := ping.NewPinger(r.Ipv4AddrIn)
	if err != nil {
		return 0, err
	}
	pinger.Count = 5
	pinger.Interval = 1 * time.Second // Adding interval between pings
	pinger.Timeout = 10 * time.Second // Increased overall timeout
	pinger.SetPrivileged(true)
	pinger.Run()
	stats := pinger.Statistics()
	return stats.MinRtt, nil
}

// RelayData represents the structure of the JSON file
type RelayData struct {
	Etag      string `json:"etag"`
	Countries []struct {
		Name   string `json:"name"`
		Code   string `json:"code"`
		Cities []struct {
			Name      string  `json:"name"`
			Code      string  `json:"code"`
			Latitude  float64 `json:"latitude"`
			Longitude float64 `json:"longitude"`
			Relays    []Relay `json:"relays"`
		} `json:"cities"`
	} `json:"countries"`
}

// ParseRelayData parses the JSON file and returns a flat list of all relays
func ParseRelayData() ([]Relay, error) {
	paths := []string{
		"/var/cache/mullvad-vpn/relays.json",
		"/Library/Caches/mullvad-vpn/relays.json",
		`C:\ProgramData\Mullvad VPN\cache\relays.json`,
	}

	var filePath string
	for _, path := range paths {
		if _, err := os.Stat(path); err == nil {
			filePath = path
			break
		}
	}

	var fileData []byte
	var err error
	if filePath == "" {
		// use embedded cache if not on filesystem
		fileData, err = mullvadclosest.GetRelays()
		if err != nil {
			return nil, fmt.Errorf("failed to read embedded file: %v", err)
		}
	} else {
		fileData, err = ioutil.ReadFile(filePath)
		if err != nil {
			return nil, fmt.Errorf("failed to read file: %v", err)
		}
	}

	var relayData RelayData
	if err := json.Unmarshal(fileData, &relayData); err != nil {
		return nil, fmt.Errorf("failed to parse JSON: %v", err)
	}

	var relays []Relay
	for _, country := range relayData.Countries {
		for _, city := range country.Cities {
			relays = append(relays, city.Relays...)
		}
	}

	return relays, nil
}
